/**
 * Class        MotorBiDir.cpp
 * Author       2020-09-27 Charles Geiser (www.dodeka.ch)
 * 
 * Purpose      Implements a class MotorBiDir which controls the speed of a brushed DC motor 
 *              by means of a PWM signal.
 *              The used dual H-Bridge motor driver L298N allows 2 brushed DC motors to be operated
 *              independently of each other at the same time in both directions of rotation.
 * 
 * Board        ESP32 Doit DevKit V1
 * Remarks      The pwm functions are taken from ESP32 ledc subsystem
 * References   https://randomnerdtutorials.com/esp32-dc-motor-l298n-motor-driver-control-speed-direction/
 * 
 * Driver                             
 *                                                  .---. 
 *                                                  |   | 
 *                                                 .-.  |
 *                                                ( M ) | Motor_A
 *                                   .----.        `-´  |
 *                                   |    |         |   |
 *                               +  .-.   |   .-----o---o------------.
 *                      6 .. 12V   ( V )  |   |  M1_A   M2_A         |
 *                               -  `-´   |   |                      |
 *                                   |    `---o 12V                  |
 *                      ESP32 GND ---+--------o GND                  |
 *                                            o 5V                   | 
 *                                            |           L298N      |
 *                                            |        Dual H-Bridge |
 *                      ESP32 18 -------------o EN_A                 |
 *                            22 -------------o IN1_A                |
 *                            23 -------------o IN2_A                |
 *                             2 -------------o IN1_B                |
 *                             4 -------------o IN2_B                |
 *                            15 -------------o EN_B                 |
 *                                            |                      |
 *                                            |  M1_B   M2_B         |
 *                                            `-----o---o------------`
 *                                                  |   |
 *                                                 .-.  |
 *                                                ( M ) | Motor_B
 *                                                 `-´  |
 *                                                  |   |
 *                                                  `---`
 *
 */
#include "MotorBiDir.h"

/**
 * Run the motor at a given speed (0..100 %)
 * 100 = full speed (no pwm Vmot = Vcc)
 * 0   = full stop  (no pwm Vmot = 0V) 
 */
void MotorBiDir::run(uint8_t speed)
{
    _percentSpeed = speed;
    _duty = map(speed, 0, 100, 0, _dutyMax);
    ledcWrite(_channel, _duty);
    rotate(_rotation);
}

/**
 * Run the motor at a given speed (0..100 %%)
 * for msToRun milliseconds
 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 * ! To be called in the loop of a state machine !
 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 */
bool MotorBiDir::runFor(uint8_t speed, uint32_t msToRun)
{
    if (_firstRun) 
    {
        //printf("%c First runFor\r\n", _id);
        _msPrevious = millis();
        _firstRun = false;
    }
    if (speed != 0) run(speed); // no need to drive motor at speed 0
    if (millis() >= _msPrevious + msToRun)
    {
        brake();
        _firstRun = true; // rearm 
    }
    return _firstRun;
}

/**
 * Do nothing for msToWait millisec
 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 * ! To be called in the loop of a state machine !
 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 */
bool MotorBiDir::waitFor(uint32_t msToWait)
{
    if (_firstRun) 
    {
        //printf("%c First waitFor\r\n", _id);
        brake();
        _msPrevious = millis();
        _firstRun = false;
    }
    if (millis() >= _msPrevious + msToWait)
    {
        _firstRun = true; // rearm
    } 
    return _firstRun;
}

/**
 * Speed up or slow down the motor from the given starting speed 
 * to the given end speed in steps of 1%.
 * 100 = full speed (no pwm Vmot = Vcc)
 * 0   = full stop  (no pwm Vmot = 0V)
 * msWait between steps defaults to 10ms but the user can supply other values
 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 * ! To be called in the loop of a state machine !
 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 */
bool MotorBiDir::accelerate(uint8_t speedFrom, uint8_t speedTo, uint32_t msWait)
{   // static variables are only initialized at very first function call
    static int8_t p = speedFrom;
    static uint8_t accelMode = 0;
    uint32_t ms;

    if (_firstRun) 
    {
        _msPrevious = millis();
        p = speedFrom;
        if (speedFrom < speedTo)  accelMode = 0; // accelerating
        if (speedFrom > speedTo)  accelMode = 1; // decelerating
        if (speedFrom == speedTo) accelMode = 2; // constant speed
        _firstRun = false;
    }
    run(p); 

    switch (accelMode)
    {
        case 0: //speed up
            if (p <= speedTo)
            {
                if ((ms = millis()) > _msPrevious + msWait)
                { 
                    _msPrevious = ms;
                    _percentSpeed = ++p;
                     //printf("%d\r\n", p);   
                }
            }
            else
            {
                _firstRun = true;
                return true;
            }      
        break;
        case 1: //slow down
            if (p >= speedTo)
            {
                if ((ms = millis()) > _msPrevious + msWait)
                { 
                    _msPrevious = ms;
                    _percentSpeed = --p;
                    //printf("%d\r\n", p);
                }
            }
            else
            {
                _firstRun = true;
                return true;
            }   
        break;
        default: // speedFrom equals speedTo --> constant speed
            _firstRun = false;
            return true;
        break;
    }
    return false;
}

/**
 * Set the rotation direction clockwise (CW) or counterclockwise (CCW)
 */
void MotorBiDir::rotate(Rotation direction)
{
    switch (direction)
    {
        case MotorBiDir::Rotation::CW:
            digitalWrite(_pin1, LOW);
            digitalWrite(_pin2, HIGH);
            _rotation = Rotation::CW;
        break;
        case MotorBiDir::Rotation::CCW:
            digitalWrite(_pin1, HIGH);
            digitalWrite(_pin2, LOW);
            _rotation = Rotation::CCW;
        break;
        default:
        break;
    }
}

/**
 * Returns the actual rotation direction
 */
MotorBiDir::Rotation MotorBiDir::rotation()
{
    return _rotation;
}

/**
 * Reverse rotation direction
 */
void MotorBiDir::reverseRotation()
{
    (_rotation == MotorBiDir::Rotation::CW) ? rotate(MotorBiDir::Rotation::CCW) : rotate(MotorBiDir::Rotation::CW);
}

/**
 * Returns the actual speed in % (0..100),
 * corresponds to duty cucle of pwm
 */
uint16_t MotorBiDir::speed()
{
    return _percentSpeed;
}

/**
 * Stop the motor immediately by connecting both 
 * motor connections to ground. This uses the back emf
 * to force the motor in the opposite direction. 
 */
void MotorBiDir::brake()
{
    digitalWrite(_pin1, LOW);
    digitalWrite(_pin2, LOW);
}

/**
 * Returns the ID of the motor
 */
char MotorBiDir::id()
{
    return _id;
}

/**
 * Prints the configuration data
 */
void MotorBiDir::printConfigData()
{
    printf("Motor ID\t%c\r\n", _id);
    printf("Pwm pin\t\t%d\r\n", _pinPwm);
    printf("IN1 pin\t\t%d\r\n", _pin1);
    printf("IN2 pin\t\t%d\r\n", _pin2);
    (rotation() == MotorBiDir::Rotation::CW) ? printf("Rotation\tCW\r\n") : printf("Rotation\tCCW\r\n");
    printf("Pwm channel\t%d\r\n", _channel);
    printf("Pwm frequency\t%d\r\n", _freq);
    printf("Pwm resolution\t%d\r\n", _resolution);
    printf("Pwm max duty\t%d\r\n\r\n", _dutyMax);
}

