/**
 * MotorBiDir.h
 * 
 * Declaration of the class MotorBiDir which controls speed and direction of rotation of a
 * brushed DC motor by means of a PWM signal. The pwm duty cycle range depending on pwm 
 * resolution is mapped to the range of 0..100 %. A speed of 0% stops the motor and a 
 * speed set to 100% runs the motor at full speed.
 * The pwm functons are taken from the ESP32 ledc subsystem.
 * 
 */
#ifndef _MotorBiDir_H_
#define _MotorBiDir_H_
#include <Arduino.h>

class MotorBiDir
{
  public:
    enum class Rotation: uint8_t { CW, CCW };
    
    MotorBiDir(char id, uint8_t pinPwmEn, uint8_t pinIn1, uint8_t pinIn2, uint8_t pwmChannel, uint32_t pwmFrequency, uint8_t pwmResolution = 8) : 
      _id(id), _pinPwm(pinPwmEn), _pin1(pinIn1), _pin2(pinIn2), _channel(pwmChannel), _freq(pwmFrequency), _resolution(pwmResolution)
    {
      ledcSetup(_channel, _freq, _resolution);
      ledcAttachPin(_pinPwm, _channel);
      ledcWrite(_channel, 0);
      _dutyMax = 1 << _resolution;
      pinMode(_pin1, OUTPUT);
      pinMode(_pin2, OUTPUT);
      digitalWrite(_pin1, LOW);   // rotate CW
      digitalWrite(_pin2, HIGH);
      _rotation = Rotation::CW;
    }

    /**
     * Run the motor at a given speed (0..100 %)
     * 100 = full speed (no pwm Vmot = Vcc)
     * 0   = full stop  (no pwm Vmot = 0V)
     */
    void run(uint8_t speed);

    /**
     * Run the motor at a given speed (0..100 %%)
     * for msToRun milliseconds
     * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     * ! To be called in the loop of a state machine !
     * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     */
    bool runFor(uint8_t speed, uint32_t msToRun);

    /**
     * Do nothing for msToWait millisec
     * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     * ! To be called in the loop of a state machine !
     * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     */
    bool waitFor(uint32_t msToWait);

    /**
     * Speed up or slow down the motor from the given starting speed 
     * to the given end speed in steps of 1%.
     * 100 = full speed (no pwm Vmot = Vcc)
     * 0   = full stop  (no pwm Vmot = 0V)
     * msWait between steps defaults to 10ms but the user can supply other values
     * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     * ! To be called in the loop of a state machine !
     * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     */
    bool accelerate(uint8_t speedFrom, uint8_t speedTo, uint32_t msWait = 10);

    /**
     * Set the rotation direction clockwise (CW) or counter clockwise (CCW)
     */
    void rotate(Rotation direction);

    /**
     * Returns the actual rotation direction (CW or CCW)
     */
    Rotation rotation();

    /**
     * Reverse direction of rotation
     */
    void reverseRotation();

    /**
     * Returns the actual speed in % (0..100),
     * corresponds to duty cucle of pwm
     */
    uint16_t speed();

    /**
     * Stop the motor immediately by connecting both 
     * motor connections to ground. This uses the back emf
     * to force the motor in the opposite direction. 
     */
    void brake();

    /**
     * Prints the configuration data
     */
    void printConfigData();

    /**
     * Returns the ID of the motor
     */
    char id();

  private:
    char _id;
    uint8_t _pinPwm;
    uint8_t _pin1;
    uint8_t _pin2;
    uint8_t _channel;
    uint32_t _freq;
    uint8_t _resolution;
    uint16_t _duty;
    uint16_t _dutyMax;
    Rotation _rotation;
    uint8_t _percentSpeed;
    uint32_t _msPrevious;
    bool _firstRun = true;
};
#endif