/**
 * Program      motorControl.cpp
 * Author       2020-09-27 Charles Geiser (https://www.dodeka.ch)
 * 
 * Purpose      Shows the usage of the class MotorBiDir to drive 2 brushed dc motors simultaneously.
 *              Each motor beforms his specific task.
 * 
 * Board        ESP32 Doit DevKit V1
 * 
 * Remarks      The pwm functions are taken from ESP32 ledc subsystem
 * 
 * References   https://randomnerdtutorials.com/esp32-dc-motor-l298n-motor-driver-control-speed-direction/
 * 
 * Driver                             
 *                                                  .---. 
 *                                                  |   | 
 *                                                 .-.  |
 *                                                ( M ) | Motor_A
 *                                   .----.        `-´  |
 *                                   |    |         |   |
 *                               +  .-.   |   .-----o---o------------.
 *                      6 .. 12V   ( V )  |   |  M1_A   M2_A         |
 *                               -  `-´   |   |                      |
 *                                   |    `---o 12V                  |
 *                      ESP32 GND ---+--------o GND                  |
 *                                        out o 5V                   | 
 *                                            |           L298N      |
 *                                            |        Dual H-Bridge |
 *                      ESP32 18 -------------o EN_A                 |
 *                            22 -------------o IN1_A                |
 *                            23 -------------o IN2_A                |
 *                             2 -------------o IN1_B                |
 *                             4 -------------o IN2_B                |
 *                            15 -------------o EN_B                 |
 *                                            |                      |
 *                                            |  M1_B   M2_B         |
 *                                            `-----o---o------------`
 *                                                  |   |
 *                                                 .-.  |
 *                                                ( M ) | Motor_B
 *                                                 `-´  |
 *                                                  |   |
 *                                                  `---`
 *
 */
#include "MotorBiDir.h"

// Configuration data for motor A
#define MOTOR_A_ID        'A'
#define PIN_PWM_EN_A      18
#define PIN_IN1_A         22
#define PIN_IN2_A         23
#define PWM_CHANNEL_A     0
#define PWM_FREQUENCY_A   210    // best value depends on motor
#define PWM_RESOLUTION_A  8

// Configuration data for motor B
#define MOTOR_B_ID        'B'
#define PIN_PWM_EN_B      15
#define PIN_IN1_B         2
#define PIN_IN2_B         4
#define PWM_CHANNEL_B     1
#define PWM_FREQUENCY_B   210    // best value depends on motor
#define PWM_RESOLUTION_B  8


MotorBiDir motorA(MOTOR_A_ID, PIN_PWM_EN_A, PIN_IN1_A, PIN_IN2_A, PWM_CHANNEL_A, PWM_FREQUENCY_A); // use default resolution 8 bit
MotorBiDir motorB(MOTOR_B_ID, PIN_PWM_EN_B, PIN_IN1_B, PIN_IN2_B, PWM_CHANNEL_B, PWM_FREQUENCY_B);

/**
 * Run the motor for msRun millisec, then stop it for msStop millisec, 
 * change direction and repeat cycle
 */
void task2A(MotorBiDir &mot, uint32_t msRun, uint32_t msStop)
{
  static uint8_t state = 0;
  switch (state)
  {
    case 0: // run motor for msRun millisec
    /**/
      if (mot.runFor(50, msRun))
      {
        state = 1;
        //printf("task2A entering state %d\r\n", state);
      }
    break;
    case 1: // stop motor for msStop millisec, then repeat cycle
     if (mot.waitFor(msStop))
      {
        state = 0;
        mot.reverseRotation();
       //printf("task2A entering state %d\r\n", state);
      }
    break;
    default:
    break;
  }
}

/**
 * Run the motor for msRun millisec, then stop it for msStop millisec, 
 * change direction and repeat cycle
 */
void task2B(MotorBiDir &mot, uint32_t msRun, uint32_t msStop)
{
  static uint8_t state = 0;

  switch (state)
  {
    case 0: // run motor for msRun millisec
    /**/
      if (mot.runFor(50, msRun))
      {
        state = 1;
        //printf("task2B entering state %d\r\n", state);
      }
    break;
    case 1: // stop motor for msStop millisec, then repeat cycle
     /**/
     if (mot.waitFor(msStop))
      {
        state = 0;
        //printf("task2B entering state %d\r\n", state);
      }
    break;
    default:
    break;
  }
}

/**
 * Accelerate the motor to speed, drive it for msRun millisec, 
 * decelerate to full stop and pause for msStop millisec
 */
void task3A(MotorBiDir &mot, uint8_t speed, uint32_t msRun, uint32_t msStop)
{
  static uint8_t state = 0;

  switch (state)
  {
    case 0:  // accelerate to speed %
      if (mot.accelerate(0, speed, 50))
      {
        state = 1;
      }    
    break;
    case 1:  // run at speed % for msRun millisec
      if (mot.runFor(speed, msRun))
      {
        state = 2;
      }
    break;
    case 2:  // decelerate from speed % to 0
      if (mot.accelerate(speed, 0, 50))
      {
        state = 3;
      }
    break;
    case 3:  // stop with brake
      mot.brake();
      if (mot.waitFor(msStop))
      {
        state = 0;
        mot.reverseRotation();
      }
    break;
    default:
    break;
  }
}

/**
 * Accelerate the motor to speed, drive it for msRun millisec, 
 * decelerate to full stop and pause for msStop millisec
 */
void task3B(MotorBiDir &mot, uint8_t speed,  uint32_t msRun, uint32_t msStop)
{
  static uint8_t state = 0;

  switch (state)
  {
    case 0: // accelerate to speed %
      if (mot.accelerate(0, speed, 50)) // takes 2.5 sec for speed = 50%
      {
        state = 1;
      }    
    break;
    case 1: // run at speed % for msRun millisec
      if (mot.runFor(speed, msRun))  // takes 1 sec (msRun)
      {
        state = 2;
      }
    break;
    case 2: // decelerate from speed % to 0
      if (mot.accelerate(speed, 0, 50)) // takes 2.5 sec for speed = 50%
      {
        state = 3;
      }
    break;
    case 3: // stop with brake
      mot.brake();
      if (mot.waitFor(msStop))  // takes 4 sec (msStop)
      {
        state = 0;
        //mot.reverseRotation();
      }
    break;
    default:
    break;
  }
}

void setup() 
{
  // No setup needed, all configuration parameters for 
  // the motors are supplied in the constructors

  motorA.printConfigData();  // show configuration data motor A
  motorB.printConfigData();  // show configuration data motor b
}

/**
 * Example tasks. Activate one task for motor A and another for motor B
 */
void loop() 
{
  task2A(motorA, 100, 900);
  //task2B(motorB, 400, 2000);
  //task3A(motorA, 100, 200, 1000);
  task3B(motorB, 50, 1000, 4000);
}